@@ -16,6 +16,7 @@ from flask_discord_interactions.context import (
 )
 
 from flask_discord_interactions.discord import (
+    InteractionType,
     DiscordInteractions,
     DiscordInteractionsBlueprint
 )
@@ -23,6 +24,13 @@ from flask_discord_interactions.discord import (
 from flask_discord_interactions.response import Response, ResponseType
 import flask_discord_interactions.embed as embed
 from flask_discord_interactions.embed import Embed
+from flask_discord_interactions.component import (
+    Component,
+    ActionRow,
+    Button,
+    ButtonStyles,
+    ComponentType
+)
 from flask_discord_interactions.client import Client
 
 
@@ -45,11 +53,17 @@ __all__ = [
     "User",
     "Role",
     "Channel",
+    "InteractionType",
     "DiscordInteractions",
     "DiscordInteractionsBlueprint",
     "Response",
     "ResponseType",
     "Embed",
+    "Component",
+    "ComponentType",
+    "ActionRow",
+    "Button",
+    "ButtonStyles",
     "Client",
 
     "InteractionResponse",
