@@ -129,6 +129,8 @@ class Member(User):
         The timestamp that the user joined the guild at.
     premium_since
         The timestamp that the user started Nitro boosting the guild at.
+    permissions
+        The permissions integer of the user.
     deaf
         Whether the user has been server deafened.
     mute
@@ -140,10 +142,15 @@ class Member(User):
     roles: list = None
     joined_at: str = None
     premium_since: str = None
+    permissions: int = None
     deaf: bool = None
     mute: bool = None
     pending: bool = None
 
+    def __post_init__(self):
+        if isinstance(self.permissions, str):
+            self.permissions = int(self.permissions)
+
     @property
     def display_name(self):
         """
@@ -215,7 +222,8 @@ class Role(ContextObject):
 @dataclass
 class Context(ContextObject):
     """
-    Represents the context in which a :class:`SlashCommand` is invoked.
+    Represents the context in which a :class:`SlashCommand` or custom ID
+    handler is invoked.
 
     Attributes
     ----------
@@ -256,15 +264,23 @@ class Context(ContextObject):
 
     client_id: str = ""
     auth_headers: dict = None
+    base_url: str = ""
+
+    custom_id: str = None
+    primary_id: str = None
+    handler_state: list = None
 
     @classmethod
     def from_data(cls, discord=None, app=None, data={}):
         if data is None:
             data = {}
 
+        custom_id = data.get("data", {}).get("custom_id")
+
         result = cls(
             client_id = app.config["DISCORD_CLIENT_ID"] if app else "",
             auth_headers = discord.auth_headers(app) if discord else {},
+            base_url = app.config["DISCORD_BASE_URL"] if app else "",
             author = Member.from_dict(data.get("member", {})),
             id = data.get("id"),
             token = data.get("token"),
@@ -272,7 +288,10 @@ class Context(ContextObject):
             guild_id = data.get("guild_id"),
             options = data.get("data", {}).get("options"),
             command_name = data.get("data", {}).get("name"),
-            command_id = data.get("data", {}).get("id")
+            command_id = data.get("data", {}).get("id"),
+            custom_id = custom_id,
+            primary_id = custom_id.split("\n", 1)[0] if custom_id else None,
+            handler_state = custom_id.split("\n") if custom_id else None
         )
 
         result.parse_resolved(data.get("data", {}).get("resolved", {}))
@@ -361,7 +380,7 @@ class Context(ContextObject):
             "@original", refers to the original message.
         """
 
-        url = ("https://discord.com/api/v8/webhooks/"
+        url = (f"{self.base_url}/webhooks/"
                f"{self.client_id}/{self.token}")
         if message is not None:
             url += f"/messages/{message}"
