@@ -1,5 +1,7 @@
 import time
 import inspect
+import uuid
+import itertools
 
 import requests
 
@@ -9,12 +11,14 @@ from nacl.exceptions import BadSignatureError
 from nacl.signing import VerifyKey
 
 from flask_discord_interactions.command import SlashCommand, SlashCommandGroup
-from flask_discord_interactions.response import ResponseType
+from flask_discord_interactions.context import Context
+from flask_discord_interactions.response import Response, ResponseType
 
 
 class InteractionType:
     PING = 1
     APPLICATION_COMMAND = 2
+    MESSAGE_COMPONENT = 3
 
 
 class DiscordInteractionsBlueprint:
@@ -25,6 +29,7 @@ class DiscordInteractionsBlueprint:
     """
     def __init__(self):
         self.discord_commands = {}
+        self.custom_id_handlers = {}
 
     def add_slash_command(self, command, name=None,
                           description=None, options=None, annotations=None):
@@ -93,6 +98,46 @@ class DiscordInteractionsBlueprint:
         self.discord_commands[name] = group
         return group
 
+    def add_custom_handler(self, handler, custom_id=None):
+        """
+        Add a handler for an incoming interaction with the specified custom ID.
+
+        Parameters
+        ----------
+        handler
+            The function to call to handle the incoming interaction.
+        custom_id
+            The custom ID to respond to. If not specified, the ID will be
+            generated randomly.
+
+        Returns
+        -------
+        str
+            The custom ID that the handler will respond to.
+        """
+        if custom_id is None:
+            custom_id = str(uuid.uuid4())
+
+        self.custom_id_handlers[custom_id] = handler
+        return custom_id
+
+    def custom_handler(self, custom_id=None):
+        """
+        Retuens a decorator to register a handler for a custom ID.
+
+        Parameters
+        ----------
+        custom_id
+            The custom ID to respond to. If not specified, the ID will be
+            generated randomly.
+        """
+        def decorator(func):
+            nonlocal custom_id
+            custom_id = self.add_custom_handler(func, custom_id)
+            return custom_id
+
+        return decorator
+
 
 class DiscordInteractions(DiscordInteractionsBlueprint):
     """
@@ -118,12 +163,14 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
             The Flask app to initialize.
         """
 
+        app.config.setdefault("DISCORD_BASE_URL", "https://discord.com/api/v9")
         app.config.setdefault("DISCORD_CLIENT_ID", "")
         app.config.setdefault("DISCORD_PUBLIC_KEY", "")
         app.config.setdefault("DISCORD_CLIENT_SECRET", "")
         app.config.setdefault("DONT_VALIDATE_SIGNATURE", False)
         app.config.setdefault("DONT_REGISTER_WITH_DISCORD", False)
         app.discord_commands = self.discord_commands
+        app.custom_id_handlers = self.custom_id_handlers
         app.discord_token = None
 
     def fetch_token(self, app=None):
@@ -151,7 +198,7 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
             app.discord_token["expires_on"] = (time.time() + app.discord_token["expires_in"]/2)
             return
         response = requests.post(
-            "https://discord.com/api/v8/oauth2/token",
+            app.config["DISCORD_BASE_URL"] + "/oauth2/token",
             data={
                 "grant_type": "client_credentials",
                 "scope": "applications.commands.update"
@@ -189,8 +236,7 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
     def update_slash_commands(self, app=None, guild_id=None):
         """
         Update the list of slash commands registered with Discord.
-        This method will delete old/unused slash commands, update modified
-        slash commands, and register new slash commands.
+        This method will overwrite all existing slash commands.
 
         Parameters
         ----------
@@ -207,57 +253,29 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
         if app.config['DONT_REGISTER_WITH_DISCORD']:
             return
 
-        needed = app.discord_commands.copy()
-
         if guild_id:
-            url = ("https://discord.com/api/v8/applications/"
+            url = (f"{app.config['DISCORD_BASE_URL']}/applications/"
                    f"{app.config['DISCORD_CLIENT_ID']}/"
                    f"guilds/{guild_id}/commands")
         else:
-            url = ("https://discord.com/api/v8/applications/"
+            url = (f"{app.config['DISCORD_BASE_URL']}/applications/"
                    f"{app.config['DISCORD_CLIENT_ID']}/commands")
 
-        response = requests.get(url, headers=self.auth_headers(app))
-        response.raise_for_status()
-        current = response.json()
+        overwrite_data = [
+            command.dump() for command in app.discord_commands.values()
+        ]
 
-        for command in current:
-            if command["name"] in needed:
-                target = needed[command["name"]]
-                if command["description"] == target.description:
-                    if (not command.get("options") and not target.options
-                            or command.get("options") == target.options):
-                        del needed[command["name"]]
+        response = requests.put(
+            url, json=overwrite_data, headers=self.auth_headers(app))
+        self.throttle(response)
 
-                        target.id = command["id"]
-                        continue
-
-            id = command["id"]
-            if guild_id:
-                delete_url = ("https://discord.com/api/v8/applications/"
-                              f"{app.config['DISCORD_CLIENT_ID']}/"
-                              f"guilds/{guild_id}/commands/{id}")
-            else:
-                delete_url = (
-                    "https://discord.com/api/v8/applications/"
-                    f"{app.config['DISCORD_CLIENT_ID']}/commands/{id}")
-
-            response = requests.delete(
-                delete_url, headers=self.auth_headers(app))
-            self.throttle(response)
+        try:
             response.raise_for_status()
-
-        for name, command in needed.items():
-            response = requests.post(
-                url, json=command.dump(), headers=self.auth_headers(app))
-            self.throttle(response)
-            try:
-                response.raise_for_status()
-            except requests.exceptions.HTTPError:
-                raise ValueError(
-                    f"Unable to register command {command.name}\n"
-                    f"{response.status_code} {response.text}")
-            command.id = response.json()["id"]
+        except requests.exceptions.HTTPError:
+            raise ValueError(
+                f"Unable to register commands:"
+                f"{response.status_code} {response.text}"
+            )
 
     def throttle(self, response):
         """
@@ -277,7 +295,7 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
         # rate_limit_bucket = response.headers["X-RateLimit-Bucket"]
 
         if not rate_limit_remaining:
-            time.sleep(rate_limit_reset - time.time())
+            time.sleep(max(rate_limit_reset - time.time(), 0))
 
     def register_blueprint(self, blueprint, app=None):
         """
@@ -299,6 +317,7 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
             app = self.app
 
         app.discord_commands.update(blueprint.discord_commands)
+        app.custom_id_handlers.update(blueprint.custom_id_handlers)
 
     def run_command(self, data):
         """
@@ -320,6 +339,53 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
 
         return slash_command.make_context_and_run(self, current_app, data)
 
+    def run_handler(self, data):
+        """
+        Run the corresponding custom ID handler given incoming interaction
+        data.
+
+        Parameters
+        ----------
+        data
+            Incoming interaction data.
+        """
+
+        context = Context.from_data(self, current_app, data)
+
+        handler = self.custom_id_handlers[context.primary_id]
+
+        args = context.handler_state[1:]
+
+        sig = inspect.signature(handler)
+
+        iterator = zip(
+            itertools.count(),
+            args,
+            itertools.islice(sig.parameters.values(), 1, None)
+        )
+
+        for i, argument, parameter in iterator:
+            annotation = parameter.annotation
+
+            if annotation == int:
+                args[i] = int(argument)
+
+            elif annotation == bool:
+                if argument == "True":
+                    args[i] = True
+                elif argument == "False":
+                    args[i] = False
+                elif argument == "None":
+                    args[i] = None
+                else:
+                    raise ValueError(
+                        f"Invalid bool in handler state parsing: {args[i]}")
+
+        result = handler(context, *args)
+
+        return Response.from_return_value(result)
+
+
     def verify_signature(self, request):
         """
         Verify the signature sent by Discord with incoming interactions.
@@ -371,10 +437,15 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
         def interactions():
             self.verify_signature(request)
 
-            if request.json.get("type") == InteractionType.PING:
+            interaction_type = request.json.get("type")
+            if interaction_type == InteractionType.PING:
                 return jsonify({"type": ResponseType.PONG})
 
-            return jsonify(self.run_command(request.json).dump())
+            elif interaction_type == InteractionType.APPLICATION_COMMAND:
+                return jsonify(self.run_command(request.json).dump())
+
+            elif interaction_type == InteractionType.MESSAGE_COMPONENT:
+                return jsonify(self.run_handler(request.json).dump_handler())
 
     def set_route_async(self, route, app=None):
         """
@@ -396,10 +467,15 @@ class DiscordInteractions(DiscordInteractionsBlueprint):
         async def interactions():
             self.verify_signature(request)
 
-            if request.json.get("type") == InteractionType.PING:
+            interaction_type = request.json.get("type")
+            if interaction_type == InteractionType.PING:
                 return jsonify({"type": ResponseType.PONG})
 
-            result = self.run_command(request.json)
+            elif interaction_type == InteractionType.APPLICATION_COMMAND:
+                result = self.run_command(request.json)
+            elif interaction_type == InteractionType.MESSAGE_COMPONENT:
+                result = self.run_handler(request.json)
+
             if inspect.isawaitable(result):
                 result = await result
 
